@@ -2,19 +2,21 @@
 """ grin searches text files.
 """
 
+import bisect
 import fnmatch
 import gzip
 import itertools
 import os
 import re
 import shlex
+import stat
 import sys
 
 import argparse
 
 
 #### Constants ####
-__version__ = '1.1.1'
+__version__ = '1.2'
 
 # Maintain the numerical order of these constants. We use them for sorting.
 PRE = -1
@@ -25,8 +27,6 @@ POST = 1
 TEXTCHARS = ''.join(map(chr, [7,8,9,10,12,13,27] + range(0x20, 0x100)))
 ALLBYTES = ''.join(map(chr, range(256)))
 
-# Color terminals; if TERM is one of these, we will attempt color output
-COLOR_TERMS = ('rxvt', 'xterm', 'xterm-color', 'linux')
 COLOR_TABLE = ['black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan',
                'white', 'default']
 COLOR_STYLE = {
@@ -37,6 +37,9 @@ COLOR_STYLE = {
 # gzip magic header bytes.
 GZIP_MAGIC = '\037\213'
 
+# Target amount of data to read into memory at a time.
+READ_BLOCKSIZE = 16 * 1024 * 1024
+
 
 def is_binary_string(bytes):
     """ Determine if a string is classified as binary rather than text.
@@ -52,22 +55,32 @@ def is_binary_string(bytes):
     nontext = bytes.translate(ALLBYTES, TEXTCHARS)
     return bool(nontext)
     
-def sliding_window(seq, n):
-    """ Returns a sliding window (up to width n) over data from the iterable
+def get_line_offsets(block):
+    """ Compute the list of offsets in DataBlock 'block' which correspond to
+    the beginnings of new lines.
 
-    Adapted from the itertools documentation.
-
-        s -> (s0,), (s0, s1), ... (s0,s1,...s[n-1]), (s1,s2,...,sn), ...
+    Returns: (offset list, count of lines in "current block")
     """
-    it = iter(seq)
-    result = ()
-    for i, elem in itertools.izip(range(n), it):
-        result += (elem,)
-        yield result
-    for elem in it:
-        result = result[1:] + (elem,)
-        yield result
-
+    # Note: this implementation based on string.find() benchmarks about twice as
+    # fast as a list comprehension using re.finditer().
+    line_offsets = [0]
+    line_count = 0    # Count of lines inside range [block.start, block.end) *only*
+    s = block.data
+    while True:
+        next_newline = s.find('\n', line_offsets[-1])
+        if next_newline < 0:
+            # Tack on a final "line start" corresponding to EOF, if not done already.
+            # This makes it possible to determine the length of each line by computing
+            # a difference between successive elements.
+            if line_offsets[-1] < len(s):
+                line_offsets.append(len(s))
+            return (line_offsets, line_count)
+        else:
+            line_offsets.append(next_newline + 1)
+            # Keep track of the count of lines within the "current block"
+            if next_newline >= block.start and next_newline < block.end:
+                line_count += 1
+            
 def colorize(s, fg=None, bg=None, bold=False, underline=False, reverse=False):
     """ Wraps a string with ANSI color escape sequences corresponding to the
     style parameters given.
@@ -130,6 +143,7 @@ def default_options():
         show_line_numbers = True,
         show_match = True,
         show_filename = True,
+        show_emacs = False,
         skip_hidden_dirs=False,
         skip_hidden_files=False,
         skip_backup_files=True,
@@ -142,6 +156,34 @@ def default_options():
     return opt
 
 
+class DataBlock(object):
+    """ This class holds a block of data read from a file, along with
+    some preceding and trailing context.
+
+    Attributes
+    ----------
+    data  : byte string
+    start : int
+        Offset into 'data' where the "current block" begins; everything
+        prior to this is 'before' context bytes
+    end : int
+        Offset into 'data' where the "current block" ends; everything
+        after this is 'after' context bytes
+    before_count : int
+        Number of lines contained in data[:start]
+    is_last : bool
+        True if this is the final block in the file
+    """
+    def __init__(self, data='', start=0, end=0, before_count=0, is_last=False):
+        self.data = data
+        self.start = start
+        self.end = end
+        self.before_count = before_count
+        self.is_last = is_last
+
+EMPTY_DATABLOCK = DataBlock()
+
+
 class GrepText(object):
     """ Grep a single file for a regex by iterating over the lines in a file.
 
@@ -154,20 +196,105 @@ class GrepText(object):
     def __init__(self, regex, options=None):
         # The compiled regex.
         self.regex = regex
+        # An equivalent regex with multiline enabled.
+        self.regex_m = re.compile(regex.pattern, regex.flags | re.MULTILINE)
 
         # The options object from parsing the configuration and command line.
         if options is None:
             options = default_options()
         self.options = options
 
+    def read_block_with_context(self, prev, fp, fp_size):
+        """ Read a block of data from the file, along with some surrounding
+        context.
+        
+        Parameters
+        ----------
+        prev : DataBlock, or None
+            The result of the previous application of read_block_with_context(),
+            or None if this is the first block.
+
+        fp : filelike object
+            The source of block data.
+        
+        fp_size : int or None
+            Size of the file in bytes, or None if the size could not be
+            determined.
+        
+        Returns
+        -------
+        A DataBlock representing the "current" block along with context.
+        """
+        if fp_size is None:
+            target_io_size = READ_BLOCKSIZE
+            block_main = fp.read(target_io_size)
+            is_last_block = len(block_main) < target_io_size
+        else:
+            remaining = max(fp_size - fp.tell(), 0)
+            target_io_size = min(READ_BLOCKSIZE, remaining)
+            block_main = fp.read(target_io_size)
+            is_last_block = target_io_size == remaining
+
+        if prev is None:
+            if is_last_block:
+                # FAST PATH: the entire file fits into a single block, so we
+                # can avoid the overhead of locating lines of 'before' and
+                # 'after' context.
+                result = DataBlock(
+                    data = block_main,
+                    start = 0,
+                    end = len(block_main),
+                    before_count = 0,
+                    is_last = True,
+                )
+                return result
+            else:
+                prev = EMPTY_DATABLOCK
+
+        # SLOW PATH: handle the general case of a large file which is split
+        # across multiple blocks.
+
+        # Look back into 'preceding' for some lines of 'before' context.
+        if prev.end == 0:
+            before_start = 0
+            before_count = 0
+        else:
+            before_start = prev.end - 1
+            before_count = 0
+            for i in range(self.options.before_context):
+                ofs = prev.data.rfind('\n', 0, before_start)
+                before_start = ofs
+                before_count += 1
+                if ofs < 0:
+                    break
+            before_start += 1
+        before_lines = prev.data[before_start:prev.end]
+        # Using readline() to force this block out to a newline boundary...
+        curr_block = (prev.data[prev.end:] + block_main +
+            ('' if is_last_block else fp.readline()))
+        # Read in some lines of 'after' context.
+        if is_last_block:
+            after_lines = ''
+        else:
+            after_lines_list = [fp.readline() for i in range(self.options.after_context)]
+            after_lines = ''.join(after_lines_list)
+
+        result = DataBlock(
+            data = before_lines + curr_block + after_lines,
+            start = len(before_lines),
+            end = len(before_lines) + len(curr_block),
+            before_count = before_count,
+            is_last = is_last_block,
+        )
+        return result
+
     def do_grep(self, fp):
         """ Do a full grep.
 
         Parameters
         ----------
         fp : filelike object
-            An open filelike object or other iterator that yields lines (with
-            line endings).
+            An open filelike object.
 
         Returns
         -------
@@ -175,44 +302,92 @@ class GrepText(object):
         each tuple of type MATCH, **spans** is a list of (start,end) positions
         of substrings that matched the pattern.
         """
-        before = self.options.before_context
-        after = self.options.after_context
-        # Start keeping track of the "after" context state. When we hit
-        # a matched line, we'll reset this counter to the number of "after"
-        # context lines we're trying to capture. For each line that follows
-        # that doesn't match, we'll decrement the counter. At 0, we'll stop
-        # accumulating "after" context lines.
-        after_state = 0
         context = []
-
-        # Make a sliding window going over the lines of the file in order to
-        # keep track of the "before" context lines.
-        window = sliding_window(fp, before + 1)
-        for i, lines in enumerate(window):
-            # The last line in the window is the line we're actually
-            # searching.
-            line = lines[-1]
-            m = self.regex.search(line)
-            if m is None:
-                if after_state != 0:
-                    # This line is part of the "after" context.
-                    context.append((i, POST, line, None))
-                    after_state -= 1
-                continue
+        line_count = 0
+        try:
+            status = os.fstat(fp.fileno())
+            if stat.S_ISREG(status.st_mode):
+                fp_size = status.st_size
             else:
-                # This line matches.
-                # Reset the "after" context counter.
-                after_state = after
-                for j, before_line in zip(range(i-len(lines)+1, i), lines[:-1]):
-                    # XXX: we can probably simply avoid adding duplicate
-                    # lines here instead of doing a second pass.
-                    context.append((j, PRE, before_line, None))
-                spans = [match.span() for match in self.regex.finditer(line)]
-                context.append((i, MATCH, line, spans))
+                fp_size = None
+        except AttributeError:  # doesn't support fileno()
+            fp_size = None
+
+        block = self.read_block_with_context(None, fp, fp_size)
+        while block.end > block.start:
+            (block_line_count, block_context) = self.do_grep_block(block,
+                    line_count - block.before_count)
+            context += block_context
+            if block.is_last:
+                break
+
+            next_block = self.read_block_with_context(block, fp, fp_size)
+            if next_block.end > next_block.start:
+                if block_line_count is None:
+                    # If the file contains N blocks, then in the best case we
+                    # will need to compute line offsets for the first N-1 blocks.
+                    # Most files will fit within a single block, so if there are
+                    # no regex matches then we can typically avoid computing *any*
+                    # line offsets.
+                    (_, block_line_count) = get_line_offsets(block)
+                line_count += block_line_count
+            block = next_block
 
         unique_context = self.uniquify_context(context)
         return unique_context
 
+    def do_grep_block(self, block, line_num_offset):
+        """ Grep a single block of file content.
+
+        Parameters
+        ----------
+        block : DataBlock
+            A chunk of file data.
+
+        line_num_offset: int
+            The number of lines preceding block.data.
+
+        Returns
+        -------
+        Tuple of the form
+            (line_count, list of (lineno, type (POST/PRE/MATCH), line, spans).
+        'line_count' is either the number of lines in the block, or None if
+        the line_count was not computed.  For each 4-tuple of type MATCH,
+        **spans** is a list of (start,end) positions of substrings that matched
+        the pattern.
+        """
+        before = self.options.before_context
+        after = self.options.after_context
+        block_context = []
+        line_offsets = None
+        line_count = None
+        
+        def build_match_context(match):
+            match_line_num = bisect.bisect(line_offsets, match.start() + block.start) - 1
+            before_count = min(before, match_line_num)
+            after_count = min(after, (len(line_offsets) - 1) - match_line_num - 1)
+            match_line = block.data[line_offsets[match_line_num]:line_offsets[match_line_num + 1]]
+            spans = [m.span() for m in self.regex.finditer(match_line)]
+
+            before_ctx = [(i + line_num_offset, PRE,
+                block.data[line_offsets[i]:line_offsets[i+1]], None)
+                    for i in range(match_line_num - before_count, match_line_num)]
+            after_ctx = [(i + line_num_offset, POST,
+                block.data[line_offsets[i]:line_offsets[i+1]], None)
+                    for i in range(match_line_num + 1, match_line_num + after_count + 1)]
+            match_ctx = [(match_line_num + line_num_offset, MATCH, match_line, spans)]
+            return before_ctx + match_ctx + after_ctx
+
+        # Using re.MULTILINE here, so ^ and $ will work as expected.
+        for match in self.regex_m.finditer(block.data[block.start:block.end]):
+            # Computing line offsets is expensive, so we do it lazily.  We don't
+            # take the extra CPU hit unless there's a regex match in the file.
+            if line_offsets is None:
+                (line_offsets, line_count) = get_line_offsets(block)
+            block_context += build_match_context(match)
+
+        return (line_count, block_context)
+
     def uniquify_context(self, context):
         """ Remove duplicate lines from the list of context lines.
         """
@@ -256,12 +431,14 @@ class GrepText(object):
             line = '%s\n' % filename
             lines.append(line)
         else:
-            if self.options.show_filename and filename is not None:
+            if self.options.show_filename and filename is not None and not self.options.show_emacs:
                 line = '%s:\n' % filename
                 if self.options.use_color:
                     line = colorize(line, **COLOR_STYLE.get('filename', {}))
                 lines.append(line)
-            if self.options.show_line_numbers:
+            if self.options.show_emacs:
+                template = '%(filename)s:%(lineno)s: %(line)s'
+            elif self.options.show_line_numbers:
                 template = '%(lineno)5s %(sep)s %(line)s'
             else:
                 template = '%(line)s'
@@ -282,6 +459,7 @@ class GrepText(object):
                     lineno = i+1,
                     sep = {PRE: '-', POST: '+', MATCH: ':'}[kind],
                     line = line,
+                    filename = filename,
                 )
                 line = template % ns
                 lines.append(line)
@@ -314,7 +492,8 @@ class GrepText(object):
             f = sys.stdin
             filename = '<STDIN>'
         else:
-            f = opener(filename, 'rb')
+            # 'r' does the right thing for both open ('rt') and gzip.open ('rb')
+            f = opener(filename, 'r')
         try:
             unique_context = self.do_grep(f)
         finally:
@@ -359,7 +538,19 @@ class FileRecognizer(object):
         self.skip_hidden_files = skip_hidden_files
         self.skip_backup_files = skip_backup_files
         self.skip_dirs = skip_dirs
-        self.skip_exts = skip_exts
+
+        # For speed, split extensions into the simple ones, that are
+        # compatible with os.path.splitext and hence can all be
+        # checked for in a single set-lookup, and the weirdos that
+        # can't and therefore must be checked for one at a time.
+        self.skip_exts_simple = set()
+        self.skip_exts_endswith = list()
+        for ext in skip_exts:
+            if os.path.splitext('foo.bar'+ext)[1] == ext:
+                self.skip_exts_simple.add(ext)
+            else:
+                self.skip_exts_endswith.append(ext)
+        
         self.skip_symlink_dirs = skip_symlink_dirs
         self.skip_symlink_files = skip_symlink_files
         self.binary_bytes = binary_bytes
@@ -391,7 +582,12 @@ class FileRecognizer(object):
         -------
         is_binary : bool
         """
-        bytes = f.read(self.binary_bytes)
+        try:
+            bytes = f.read(self.binary_bytes)
+        except Exception, e:
+            # When trying to read from something that looks like a gzipped file,
+            # it may be corrupt. If we do get an error, assume that the file is binary.
+            return True
         return is_binary_string(bytes)
 
     def is_gzipped_text(self, filename):
@@ -461,10 +657,19 @@ class FileRecognizer(object):
         -------
         kind : str
         """
-        if os.path.isdir(filename):
-            return self.recognize_directory(filename)
-        else:
-            return self.recognize_file(filename)
+        try:
+            st_mode = os.stat(filename).st_mode
+            if stat.S_ISREG(st_mode):
+                return self.recognize_file(filename)
+            elif stat.S_ISDIR(st_mode):
+                return self.recognize_directory(filename)
+            else:
+                # We're only interested in regular files and directories.
+                # A named pipe in particular would be problematic, because
+                # it would cause open() to hang indefinitely.
+                return 'skip'
+        except OSError:
+            return 'unreadable'
         
     def recognize_directory(self, filename):
         """ Determine what to do with a directory.
@@ -477,13 +682,7 @@ class FileRecognizer(object):
             return 'link'
         if basename in self.skip_dirs:
             return 'skip'
-        # Follow any possible symlink to the real file in order to check its
-        # permissions.
-        filename = os.path.realpath(filename)
-        if os.access(filename, os.R_OK|os.X_OK):
-            return 'directory'
-
-        return 'unreadable'
+        return 'directory'
 
     def recognize_file(self, filename):
         """ Determine what to do with a file.
@@ -495,26 +694,24 @@ class FileRecognizer(object):
             return 'skip'
         if self.skip_symlink_files and os.path.islink(filename):
             return 'link'
-        base, ext = os.path.splitext(filename)
-        if ext in self.skip_exts:
+        
+        filename_nc = os.path.normcase(filename)
+        ext = os.path.splitext(filename_nc)[1]
+        if ext in self.skip_exts_simple or ext.startswith('.~'):
             return 'skip'
-        # Follow any possible symlink to the real file in order to check its
-        # permissions.
-        filename = os.path.realpath(filename)
-        if os.access(filename, os.R_OK):
-            # Just to be sure, catch OSErrors.
-            try:
-                if self.is_binary(filename):
-                    if self.is_gzipped_text(filename):
-                        return 'gzip'
-                    else:
-                        return 'binary'
+        for ext in self.skip_exts_endswith:
+            if filename_nc.endswith(ext):
+                return 'skip'
+        try:
+            if self.is_binary(filename):
+                if self.is_gzipped_text(filename):
+                    return 'gzip'
                 else:
-                    return 'text'
-            except OSError:
-                return 'unreadable'
-
-        return 'unreadable'
+                    return 'binary'
+            else:
+                return 'text'
+        except (OSError, IOError):
+            return 'unreadable'
 
     def walk(self, startpath):
         """ Walk the tree from a given start path yielding all of the files (not
@@ -538,7 +735,10 @@ class FileRecognizer(object):
             # Not a directory, so there is no need to recurse.
             return
         elif kind == 'directory':
-            basenames = os.listdir(startpath)
+            try:
+                basenames = os.listdir(startpath)
+            except OSError:
+                return
             for basename in basenames:
                 path = os.path.join(startpath, basename)
                 for fn, k in self.walk(path):
@@ -552,10 +752,11 @@ def get_grin_arg_parser(parser=None):
         parser = argparse.ArgumentParser(
             description="Search text files for a given regex pattern.",
             epilog="Bug reports to <enthought-dev@mail.enthought.com>.",
-            version='grin %s' % __version__,
             formatter_class=argparse.RawDescriptionHelpFormatter,
         )
 
+    parser.add_argument('-v', '--version', action='version', version='grin %s' % __version__,
+        help="show program's version number and exit")
     parser.add_argument('-i', '--ignore-case', action='append_const',
         dest='re_flags', const=re.I, default=[], help="ignore case in the regex")
     parser.add_argument('-A', '--after-context', default=0, type=int,
@@ -577,6 +778,9 @@ def get_grin_arg_parser(parser=None):
     parser.add_argument('--without-filename', action='store_false',
         dest='show_filename', 
         help="do not show the filenames of files that match")
+    parser.add_argument('--emacs', action='store_true',
+        dest='show_emacs',
+        help="print the filename with every match for easier parsing by e.g. Emacs")
     parser.add_argument('-l', '--files-with-matches', action='store_false',
         dest='show_match',
         help="show only the filenames and not the texts of the matches")
@@ -598,10 +802,10 @@ def get_grin_arg_parser(parser=None):
         help="do skip .hidden files [default]")
     parser.add_argument('-b', '--no-skip-backup-files',
         dest='skip_backup_files', action='store_false',
-        help="do not skip backup~ files")
+        help="do not skip backup~ files [deprecated; edit --skip-exts]")
     parser.add_argument('--skip-backup-files',
         dest='skip_backup_files', action='store_true', default=True,
-        help="do skip backup~ files [default]")
+        help="do skip backup~ files [default] [deprecated; edit --skip-exts]")
     parser.add_argument('-S', '--no-skip-hidden-dirs', dest='skip_hidden_dirs',
         action='store_false',
         help="do not skip .hidden directories")
@@ -615,15 +819,22 @@ def get_grin_arg_parser(parser=None):
         action='store_const', const='',
         help="do not skip any directories")
     parser.add_argument('-e', '--skip-exts',
-        default='.pyc,.pyo,.so,.o,.a,.tgz',
+        default='.pyc,.pyo,.so,.o,.a,.tgz,.tar.gz,.rar,.zip,~,#,.bak,.png,.jpg,.gif,.bmp,.tif,.tiff,.pyd,.dll,.exe,.obj,.lib',
         help="comma-separated list of file extensions to skip [default=%(default)r]")
     parser.add_argument('-E', '--no-skip-exts', dest='skip_exts',
         action='store_const', const='',
         help="do not skip any file extensions")
+    parser.add_argument('--no-follow', action='store_false', dest='follow_symlinks',
+        default=False,
+        help="do not follow symlinks to directories and files [default]")
+    parser.add_argument('--follow', action='store_true', dest='follow_symlinks',
+        help="follow symlinks to directories and files")
     parser.add_argument('-f', '--files-from-file', metavar="FILE",
         help="read files to search from a file, one per line; - for stdin")
     parser.add_argument('-0', '--null-separated', action='store_true',
         help="filenames specified in --files-from-file are separated by NULs")
+    parser.add_argument('--sys-path', action='store_true',
+        help="search the directories on sys.path")
 
     parser.add_argument('regex', help="the regular expression to search for")
     parser.add_argument('files', nargs='*', help="the files to search")
@@ -637,9 +848,10 @@ def get_grind_arg_parser(parser=None):
         parser = argparse.ArgumentParser(
             description="Find text and binary files using similar rules as grin.",
             epilog="Bug reports to <enthought-dev@mail.enthought.com>.",
-            version='grind %s' % __version__,
         )
 
+    parser.add_argument('-v', '--version', action='version', version='grin %s' % __version__,
+        help="show program's version number and exit")
     parser.add_argument('-s', '--no-skip-hidden-files',
         dest='skip_hidden_files', action='store_false',
         help="do not skip .hidden files")
@@ -648,10 +860,10 @@ def get_grind_arg_parser(parser=None):
         help="do skip .hidden files")
     parser.add_argument('-b', '--no-skip-backup-files',
         dest='skip_backup_files', action='store_false',
-        help="do not skip backup~ files")
+        help="do not skip backup~ files [deprecated; edit --skip-exts]")
     parser.add_argument('--skip-backup-files',
         dest='skip_backup_files', action='store_true', default=True,
-        help="do skip backup~ files")
+        help="do skip backup~ files [default] [deprecated; edit --skip-exts]")
     parser.add_argument('-S', '--no-skip-hidden-dirs', dest='skip_hidden_dirs',
         action='store_false',
         help="do not skip .hidden directories")
@@ -665,15 +877,22 @@ def get_grind_arg_parser(parser=None):
         action='store_const', const='',
         help="do not skip any directories")
     parser.add_argument('-e', '--skip-exts',
-        default='.pyc,.pyo,.so,.o,.a,.tgz',
+        default='.pyc,.pyo,.so,.o,.a,.tgz,.tar.gz,.rar,.zip,~,#,.bak,.png,.jpg,.gif,.bmp,.tif,.tiff,.pyd,.dll,.exe,.obj,.lib',
         help="comma-separated list of file extensions to skip [default=%(default)r]")
     parser.add_argument('-E', '--no-skip-exts', dest='skip_exts',
         action='store_const', const='',
         help="do not skip any file extensions")
+    parser.add_argument('--no-follow', action='store_false', dest='follow_symlinks',
+        default=False,
+        help="do not follow symlinks to directories and files [default]")
+    parser.add_argument('--follow', action='store_true', dest='follow_symlinks',
+        help="follow symlinks to directories and files")
     parser.add_argument('-0', '--null-separated', action='store_true',
         help="print the filenames separated by NULs")
     parser.add_argument('--dirs', nargs='+', default=["."],
         help="the directories to start from")
+    parser.add_argument('--sys-path', action='store_true',
+        help="search the directories on sys.path")
 
     parser.add_argument('glob', default='*', nargs='?',
         help="the glob pattern to match; you may need to quote this to prevent "
@@ -690,6 +909,8 @@ def get_recognizer(args):
         skip_hidden_dirs=args.skip_hidden_dirs,
         skip_dirs=set(args.skip_dirs.split(',')),
         skip_exts=set(args.skip_exts.split(',')),
+        skip_symlink_files=not args.follow_symlinks,
+        skip_symlink_dirs=not args.follow_symlinks,
     )
     return fr
 
@@ -706,6 +927,10 @@ def get_filenames(args):
     filename : str
     kind : either 'text' or 'gzip'
         What kind of file it is.
+
+    Raises
+    ------
+    IOError if a requested file cannot be found.
     """
     files = []
     # If the user has given us a file with filenames, consume them first.
@@ -716,9 +941,14 @@ def get_filenames(args):
         elif os.path.exists(args.files_from_file):
             files_file = open(args.files_from_file)
             should_close = True
+        else:
+            raise IOError(2, 'No such file: %r' % args.files_from_file)
 
         try:
             # Remove ''
+            # XXX: how can I detect bad filenames? One user accidentally ran
+            # grin -f against a binary file and got an unhelpful error message
+            # later.
             if args.null_separated:
                 files.extend([x.strip() for x in files_file.read().split('\0')])
             else:
@@ -729,13 +959,22 @@ def get_filenames(args):
 
     # Now add the filenames provided on the command line itself.
     files.extend(args.files)
+    if args.sys_path:
+        files.extend(sys.path)
+    # Make sure we don't have any empty strings lying around.
+    # Also skip certain special null files which may be added by programs like
+    # Emacs.
+    if sys.platform == 'win32':
+        upper_bad = set(['NUL:', 'NUL'])
+        raw_bad = set([''])
+    else:
+        upper_bad = set()
+        raw_bad = set(['', '/dev/null'])
+    files = [fn for fn in files if fn not in raw_bad and fn.upper() not in upper_bad]
     if len(files) == 0:
         # Add the current directory at least.
         files = ['.']
 
-    # Make sure we don't have any empty strings lying around.
-    files = [fn for fn in files if fn != '']
-
     # Go over our list of filenames and see if we can recognize each as
     # something we want to grep.
     fr = get_recognizer(args)
@@ -765,25 +1004,28 @@ def get_regex(args):
 
 
 def grin_main(argv=None):
-    if argv is None:
-        # Look at the GRIN_ARGS environment variable for more arguments.
-        env_args = shlex.split(os.getenv('GRIN_ARGS', ''))
-        argv = [sys.argv[0]] + env_args + sys.argv[1:]
-    parser = get_grin_arg_parser()
-    args = parser.parse_args(argv[1:])
-    if args.context is not None:
-        args.before_context = args.context
-        args.after_context = args.context
-    args.use_color = args.force_color or (not args.no_color and
-        sys.stdout.isatty() and
-        (os.environ.get('TERM') in COLOR_TERMS))
+    try:
+        if argv is None:
+            # Look at the GRIN_ARGS environment variable for more arguments.
+            env_args = shlex.split(os.getenv('GRIN_ARGS', ''))
+            argv = [sys.argv[0]] + env_args + sys.argv[1:]
+        parser = get_grin_arg_parser()
+        args = parser.parse_args(argv[1:])
+        if args.context is not None:
+            args.before_context = args.context
+            args.after_context = args.context
+        args.use_color = args.force_color or (not args.no_color and
+            sys.stdout.isatty() and
+            (os.environ.get('TERM') != 'dumb'))
 
-    regex = get_regex(args)
-    g = GrepText(regex, args)
-    openers = dict(text=open, gzip=gzip.open)
-    for filename, kind in get_filenames(args):
-        report = g.grep_a_file(filename, opener=openers[kind])
-        sys.stdout.write(report)
+        regex = get_regex(args)
+        g = GrepText(regex, args)
+        openers = dict(text=open, gzip=gzip.open)
+        for filename, kind in get_filenames(args):
+            report = g.grep_a_file(filename, opener=openers[kind])
+            sys.stdout.write(report)
+    except KeyboardInterrupt:
+        raise SystemExit(0)
 
 def print_line(filename):
     print filename
@@ -795,24 +1037,30 @@ def print_null(filename):
     sys.stdout.write('\0')
 
 def grind_main(argv=None):
-    if argv is None:
-        # Look at the GRIND_ARGS environment variable for more arguments.
-        env_args = shlex.split(os.getenv('GRIND_ARGS', ''))
-        argv = [sys.argv[0]] + env_args + sys.argv[1:]
-    parser = get_grind_arg_parser()
-    args = parser.parse_args(argv[1:])
+    try:
+        if argv is None:
+            # Look at the GRIND_ARGS environment variable for more arguments.
+            env_args = shlex.split(os.getenv('GRIND_ARGS', ''))
+            argv = [sys.argv[0]] + env_args + sys.argv[1:]
+        parser = get_grind_arg_parser()
+        args = parser.parse_args(argv[1:])
 
-    # Define the output function.
-    if args.null_separated:
-        output = print_null
-    else:
-        output = print_line
-        
-    fr = get_recognizer(args)
-    for dir in args.dirs:
-        for filename, k in fr.walk(dir):
-            if fnmatch.fnmatch(os.path.basename(filename), args.glob):
-                output(filename)
+        # Define the output function.
+        if args.null_separated:
+            output = print_null
+        else:
+            output = print_line
+
+        if args.sys_path:
+            args.dirs.extend(sys.path)
+
+        fr = get_recognizer(args)
+        for dir in args.dirs:
+            for filename, k in fr.walk(dir):
+                if fnmatch.fnmatch(os.path.basename(filename), args.glob):
+                    output(filename)
+    except KeyboardInterrupt:
+        raise SystemExit(0)
 
 if __name__ == '__main__':
     grin_main()
