@@ -4,6 +4,7 @@
 import gzip
 import os
 import shutil
+import socket
 import sys
 
 import nose
@@ -47,6 +48,10 @@ def binary_middle(filename, open=open):
     f.write(text)
     f.close()
 
+def socket_file(filename):
+    s = socket.socket(socket.AF_UNIX)
+    s.bind(filename)
+
 def unreadable_file(filename):
     """ Write a file that does not have read permissions.
     """
@@ -78,6 +83,8 @@ def setup():
     binary_middle('binary_middle')
     text_file('text')
     text_file('text~')
+    text_file('text#')
+    text_file('foo.bar.baz')
     os.mkdir('dir')
     binary_file('.binary')
     text_file('.text')
@@ -107,6 +114,7 @@ def setup():
     text_file('text.dont_skip_ext')
     os.mkdir('skip_dir')
     text_file('fake_skip_dir')
+    socket_file('socket_test')
 
     # Make a directory tree to test tree-walking.
     os.mkdir('tree')
@@ -152,7 +160,7 @@ def teardown():
         'unreadable_dir', 'unexecutable_dir', 'totally_unusable_dir',
         'unreadable_file_link', 'unreadable_dir_link', 'unexecutable_dir_link',
         'totally_unusable_dir_link', 'text.skip_ext', 'text.dont_skip_ext',
-        'dir.skip_ext', 'skip_dir', 'fake_skip_dir',
+        'dir.skip_ext', 'skip_dir', 'fake_skip_dir', 'text#', 'foo.bar.baz',
     ]
     for filename in files_to_delete:
         try:
@@ -162,7 +170,7 @@ def teardown():
                 os.rmdir(filename)
         except Exception, e:
             print >>sys.stderr, 'Could not delete %s: %s' % (filename, e)
-
+    os.unlink('socket_test')
     os.path.walk('tree', ensure_deletability, None)
     shutil.rmtree('tree')
 
@@ -201,6 +209,10 @@ def test_binary_middle():
     assert fr.recognize_file('binary_middle') == 'binary'
     assert fr.recognize('binary_middle') == 'binary'
 
+def test_socket():
+    fr= FileRecognizer()
+    assert fr.recognize('socket_test') == 'skip'
+
 def test_dir():
     fr = FileRecognizer()
     assert fr.recognize_directory('dir') == 'directory'
@@ -251,6 +263,14 @@ def test_do_not_skip_backup():
     fr = FileRecognizer(skip_backup_files=False)
     assert fr.recognize_file('text~') == 'text'
 
+def test_skip_weird_exts():
+    fr = FileRecognizer(skip_exts=set())
+    assert fr.recognize_file('text#') == 'text'
+    assert fr.recognize_file('foo.bar.baz') == 'text'
+    fr = FileRecognizer(skip_exts=set(['#', '.bar.baz']))
+    assert fr.recognize_file('text#') == 'skip'
+    assert fr.recognize_file('foo.bar.baz') == 'skip'
+
 def test_do_not_skip_hidden_or_symlinks():
     fr = FileRecognizer(skip_hidden_files=False, skip_hidden_dirs=False,
         skip_symlink_dirs=False, skip_symlink_files=False)
@@ -295,23 +315,23 @@ def test_lack_of_permissions():
     fr = FileRecognizer()
     assert fr.recognize('unreadable_file') == 'unreadable'
     assert fr.recognize_file('unreadable_file') == 'unreadable'
-    assert fr.recognize('unreadable_directory') == 'unreadable'
-    assert fr.recognize_directory('unreadable_dir') == 'unreadable'
-    assert fr.recognize('unexecutable_dir') == 'unreadable'
-    assert fr.recognize_directory('unexecutable_dir') == 'unreadable'
-    assert fr.recognize('totally_unusable_dir') == 'unreadable'
-    assert fr.recognize_directory('totally_unusable_dir') == 'unreadable'
+    assert fr.recognize('unreadable_dir') == 'directory'
+    assert fr.recognize_directory('unreadable_dir') == 'directory'
+    assert fr.recognize('unexecutable_dir') == 'directory'
+    assert fr.recognize_directory('unexecutable_dir') == 'directory'
+    assert fr.recognize('totally_unusable_dir') == 'directory'
+    assert fr.recognize_directory('totally_unusable_dir') == 'directory'
 
 def test_symlink_src_unreadable():
     fr = FileRecognizer(skip_symlink_files=False, skip_symlink_dirs=False)
     assert fr.recognize('unreadable_file_link') == 'unreadable'
     assert fr.recognize_file('unreadable_file_link') == 'unreadable'
-    assert fr.recognize('unreadable_directory_link') == 'unreadable'
-    assert fr.recognize_directory('unreadable_dir_link') == 'unreadable'
-    assert fr.recognize('unexecutable_dir_link') == 'unreadable'
-    assert fr.recognize_directory('unexecutable_dir_link') == 'unreadable'
-    assert fr.recognize('totally_unusable_dir_link') == 'unreadable'
-    assert fr.recognize_directory('totally_unusable_dir_link') == 'unreadable'
+    assert fr.recognize('unreadable_dir_link') == 'directory'
+    assert fr.recognize_directory('unreadable_dir_link') == 'directory'
+    assert fr.recognize('unexecutable_dir_link') == 'directory'
+    assert fr.recognize_directory('unexecutable_dir_link') == 'directory'
+    assert fr.recognize('totally_unusable_dir_link') == 'directory'
+    assert fr.recognize_directory('totally_unusable_dir_link') == 'directory'
 
 def test_skip_ext():
     fr = FileRecognizer(skip_exts=set(['.skip_ext']))
